#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>
#include "BoardP.h"

#include "pieces.bm"

static XtResource resources[] = {
#define offset(field) XtOffsetOf(BoardRec, board.field)
    /* {name, class, type, size, offset, default_type, default_addr}, */
    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(foreground), XtRString, XtDefaultForeground },
    { XtNblackForeground, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(black_fg), XtRString, XtDefaultBackground },
    { XtNblackBackground, XtCBackground, XtRPixel, sizeof(Pixel),
	offset(black_bg), XtRString, XtDefaultForeground },
    { XtNblackBorder, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(black_bd), XtRString, XtDefaultBackground },
    { XtNredForeground, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(red_fg), XtRString, XtDefaultForeground },
    { XtNredBackground, XtCBackground, XtRPixel, sizeof(Pixel),
	offset(red_bg), XtRString, XtDefaultBackground },
    { XtNredBorder, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(red_bd), XtRString, XtDefaultBackground },
    /*{ XtNboardFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(font), XtRString, XtDefaultFont },*/
    { XtNlineWidth, XtCLineWidth, XtRInt, sizeof(int),
	offset(line_width), XtRImmediate, (XtPointer)1 },
    { XtNborderWidth, XtCBorderWidth, XtRInt, sizeof(int),
	offset(border_width), XtRImmediate, (XtPointer)1 },
    { XtNside, XtCSide, XtRInt, sizeof(int),
	offset(side), XtRImmediate, 0 },
    { XtNturn, XtCTurn, XtRInt, sizeof(int),
	offset(turn), XtRImmediate, 0 },
    { XtNblinkInterval, XtCBlinkInterval, XtRInt, sizeof(int),
	offset(blink_interval), XtRImmediate, (XtPointer)1000 },
    { XtNmovedCallback, XtCMovedCallback, XtRCallback, sizeof(XtPointer),
	offset(moved_callbacks), XtRCallback, NULL },
#undef offset
};

static void Blink(), BlinkOn(), BlinkOff();
static void StartMove(/* Widget, XEvent*, String*, Cardinal* */);

static XtActionsRec actions[] =
{
  /* {name, procedure}, */
    {"Move",	StartMove},
};

static char translations[] =
"<Btn1Down>:	Move()\n\
";

static void	Initialize();
static void	Destroy();
static void	Redisplay();
static void	Resize();
static Boolean	SetValues();

BoardClassRec boardClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &coreClassRec,
    /* class_name		*/	"Board",
    /* widget_size		*/	sizeof(BoardRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	XtInheritRealize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* board fields */
    /* empty			*/	0
  }
};

WidgetClass boardWidgetClass = (WidgetClass)&boardClassRec;

static void DrawPieces(), DrawGrid();

#define	min(x,y)	((x) < (y) ? (x) : (y))
#define	max(x,y)	((x) > (y) ? (x) : (y))

static void
init_size(w)
    BoardWidget w;
{
    Dimension	wd, ht, spacing;

    if (w->core.height <= 0 && w->core.width <= 0) {
	spacing = 80;
	w->core.height = spacing*(ROWS+1);
	w->core.width = spacing*(COLS+1);
	w->board.origin_x = spacing;
	w->board.origin_y = spacing;
    } else {
	wd = w->core.width;
	ht = w->core.height;
	/*spacing = min(wd/ROWS, ht/COLS);*/
	spacing = 80;
	w->board.origin_x = (wd - spacing*(COLS-1))/2;
	w->board.origin_y = (ht - spacing*(ROWS-1))/2;
    }
    w->board.spacing = spacing;
    w->board.radius = 36;
}

static void
init_pieces(w)
    BoardWidget w;
{
    int i, j;

    w->board.turn = RED;
    w->board.blink_pc = EMPTY;
    w->board.blink_show = TRUE;
    w->board.blink_on = FALSE;

    for (i = 0; i < COLS; i++)
    for (j = 0; j < ROWS; j++)
	w->board.pieces[i][j] = EMPTY;

    if (w->board.side == BLACK) {
	w->board.pieces[0][0] = BJU;
	w->board.pieces[1][0] = BMA;
	w->board.pieces[2][0] = BXIANG;
	w->board.pieces[3][0] = BSHI;
	w->board.pieces[4][0] = BJIANG;
	w->board.pieces[5][0] = BSHI;
	w->board.pieces[6][0] = BXIANG;
	w->board.pieces[7][0] = BMA;
	w->board.pieces[8][0] = BJU;

	w->board.pieces[1][2] = BPAO;
	w->board.pieces[7][2] = BPAO;

	w->board.pieces[0][3] = BZU;
	w->board.pieces[2][3] = BZU;
	w->board.pieces[4][3] = BZU;
	w->board.pieces[6][3] = BZU;
	w->board.pieces[8][3] = BZU;

	w->board.pieces[0][ROWS-1] = RJU;
	w->board.pieces[1][ROWS-1] = RMA;
	w->board.pieces[2][ROWS-1] = RXIANG;
	w->board.pieces[3][ROWS-1] = RSHI;
	w->board.pieces[4][ROWS-1] = RSHUAI;
	w->board.pieces[5][ROWS-1] = RSHI;
	w->board.pieces[6][ROWS-1] = RXIANG;
	w->board.pieces[7][ROWS-1] = RMA;
	w->board.pieces[8][ROWS-1] = RJU;

	w->board.pieces[1][ROWS-3] = RPAO;
	w->board.pieces[7][ROWS-3] = RPAO;

	w->board.pieces[0][ROWS-4] = RBING;
	w->board.pieces[2][ROWS-4] = RBING;
	w->board.pieces[4][ROWS-4] = RBING;
	w->board.pieces[6][ROWS-4] = RBING;
	w->board.pieces[8][ROWS-4] = RBING;
    } else {
	w->board.pieces[0][0] = RJU;
	w->board.pieces[1][0] = RMA;
	w->board.pieces[2][0] = RXIANG;
	w->board.pieces[3][0] = RSHI;
	w->board.pieces[4][0] = RSHUAI;
	w->board.pieces[5][0] = RSHI;
	w->board.pieces[6][0] = RXIANG;
	w->board.pieces[7][0] = RMA;
	w->board.pieces[8][0] = RJU;

	w->board.pieces[1][2] = RPAO;
	w->board.pieces[7][2] = RPAO;

	w->board.pieces[0][3] = RBING;
	w->board.pieces[2][3] = RBING;
	w->board.pieces[4][3] = RBING;
	w->board.pieces[6][3] = RBING;
	w->board.pieces[8][3] = RBING;

	w->board.pieces[0][ROWS-1] = BJU;
	w->board.pieces[1][ROWS-1] = BMA;
	w->board.pieces[2][ROWS-1] = BXIANG;
	w->board.pieces[3][ROWS-1] = BSHI;
	w->board.pieces[4][ROWS-1] = BJIANG;
	w->board.pieces[5][ROWS-1] = BSHI;
	w->board.pieces[6][ROWS-1] = BXIANG;
	w->board.pieces[7][ROWS-1] = BMA;
	w->board.pieces[8][ROWS-1] = BJU;

	w->board.pieces[1][ROWS-3] = BPAO;
	w->board.pieces[7][ROWS-3] = BPAO;

	w->board.pieces[0][ROWS-4] = BZU;
	w->board.pieces[2][ROWS-4] = BZU;
	w->board.pieces[4][ROWS-4] = BZU;
	w->board.pieces[6][ROWS-4] = BZU;
	w->board.pieces[8][ROWS-4] = BZU;
    }
}

static void
init_pixmaps(w)
    BoardWidget w;
{
    Display *dpy = XtDisplay(w);
    Window win = RootWindowOfScreen(XtScreen(w));
    int dp = DefaultDepthOfScreen(XtScreen(w));
    Dimension wd = w->core.width;
    Dimension ht = w->core.height;
    Dimension d = w->board.radius*2;

    w->board.board_pixmap = XCreatePixmap(dpy, win, wd, ht, dp);
    w->board.grid_pixmap = XCreatePixmap(dpy, win, wd, ht, 1);
    w->board.temp_pixmap = XCreatePixmap(dpy, win, wd, ht, dp);
    w->board.mask_pixmap = XCreatePixmap(dpy, win, d, d, 1);
}

static void
load_pieces(w)
    BoardWidget w;
{
    Display *dpy = XtDisplay(w);
    Window win = RootWindowOfScreen(XtScreen(w));
    Dimension d = w->board.radius*2;
    Pixmap bm, pm;
    int i;

    for (i = BJU; i <= BZU; i++) {
	bm = XCreatePixmapFromBitmapData(dpy, win, pieces_bits[i-BJU],
	    48, 48, 1, 0, 1);
	pm = XCreatePixmap(dpy, win, d, d, 1);
	XFillRectangle(dpy, pm, w->board.bg_gc, 0, 0, d, d);
	if (w->board.black_bg == w->core.background_pixel) {
	    XDrawArc(dpy, pm, w->board.fg_gc, 0, 0, d, d, 0, 360*64);
	    XDrawArc(dpy, pm, w->board.fg_gc, 1, 1, d-2, d-2, 0, 360*64);
	    XDrawArc(dpy, pm, w->board.fg_gc, 2, 2, d-4, d-4, 0, 360*64);
	}
	XCopyArea(dpy, bm, pm, w->board.fg_gc, 0, 0, 48, 48, d/2-24, d/2-24);
	w->board.pieces_pixmap[i-BJU] = pm;
	XFreePixmap(dpy, bm);
    }

    for (i = RJU; i <= RBING; i++) {
	bm = XCreatePixmapFromBitmapData(dpy, win, pieces_bits[i-BJU],
	    48, 48, 1, 0, 1);
	pm = XCreatePixmap(dpy, win, d, d, 1);
	XFillRectangle(dpy, pm, w->board.bg_gc, 0, 0, d, d);
	if (w->board.red_bg == w->core.background_pixel) {
	    XDrawArc(dpy, pm, w->board.fg_gc, 0, 0, d, d, 0, 360*64);
	    XDrawArc(dpy, pm, w->board.fg_gc, 1, 1, d-2, d-2, 0, 360*64);
	    XDrawArc(dpy, pm, w->board.fg_gc, 2, 2, d-4, d-4, 0, 360*64);
	}
	XCopyArea(dpy, bm, pm, w->board.fg_gc, 0, 0, 48, 48, d/2-24, d/2-24);
	w->board.pieces_pixmap[i-BJU] = pm;
	XFreePixmap(dpy, bm);
    }
}

static void
init_gcs(w)
    BoardWidget w;
{
    Display *dpy = XtDisplay(w);
    Dimension d = w->board.radius*2;
    XGCValues values;
    XtGCMask mask;

    /*mask = GCForeground | GCBackground | GCLineWidth | GCFont;*/
    mask = GCForeground | GCBackground | GCLineWidth;
    values.foreground = 1;
    values.background = 0;
    values.line_width = w->board.line_width;
    /*values.font = w->board.font->fid;*/
    w->board.fg_gc = XCreateGC(dpy, w->board.grid_pixmap, mask, &values);

    mask = GCForeground | GCBackground;
    values.foreground = 0;
    values.background = 1;
    w->board.bg_gc = XCreateGC(dpy, w->board.grid_pixmap, mask, &values);

    mask = GCForeground | GCBackground;
    values.foreground = w->board.foreground;
    values.background = w->core.background_pixel;
    w->board.copy_gc = XCreateGC(dpy, w->board.board_pixmap, mask, &values);

    XFillRectangle(dpy, w->board.mask_pixmap, w->board.bg_gc, 0, 0, d, d);
    XFillArc(dpy, w->board.mask_pixmap, w->board.fg_gc, 0, 0, d, d, 0, 360*64);

    mask = GCClipMask | GCForeground | GCBackground;
    values.clip_mask = w->board.mask_pixmap;
    values.foreground = w->board.black_fg;
    values.background = w->board.black_bg;
    w->board.black_gc = XCreateGC(dpy, w->board.board_pixmap, mask, &values);

    mask = GCClipMask | GCForeground | GCBackground;
    values.clip_mask = w->board.mask_pixmap;
    values.foreground = w->board.red_fg;
    values.background = w->board.red_bg;
    w->board.red_gc = XCreateGC(dpy, w->board.board_pixmap, mask, &values);
}

static void
Initialize(req, new)
    BoardWidget req, new;
{
    init_size(new);
    init_pieces(new);
    init_pixmaps(new);
    init_gcs(new);
    load_pieces(new);

    new->board.move_cursor = XCreateFontCursor(XtDisplay(new), XC_fleur);

    DrawGrid(new);
    DrawPieces(new);
}

static void
destroy_pixmaps(w)
    BoardWidget w;
{
    int i;

    XFreePixmap(XtDisplay(w), w->board.board_pixmap);
    XFreePixmap(XtDisplay(w), w->board.grid_pixmap);
    XFreePixmap(XtDisplay(w), w->board.temp_pixmap);
    XFreePixmap(XtDisplay(w), w->board.mask_pixmap);

    for (i = BJU; i <= RBING; i++)
	XFreePixmap(XtDisplay(w), w->board.pieces_pixmap[i-BJU]);
}

static void
destroy_gcs(w)
    BoardWidget w;
{
    XFreeGC(XtDisplay(w), w->board.fg_gc);
    XFreeGC(XtDisplay(w), w->board.bg_gc);
    XFreeGC(XtDisplay(w), w->board.copy_gc);
    XFreeGC(XtDisplay(w), w->board.black_gc);
    XFreeGC(XtDisplay(w), w->board.red_gc);
}

static void
Destroy(w)
    BoardWidget w;
{
    destroy_gcs(w);
    destroy_pixmaps(w);
}

static void
Redisplay(w, event, region)
    BoardWidget w;
    XExposeEvent *event;
    Region region;
{
    Position x, y;
    Dimension wd, ht;

    x = event->x;
    y = event->y;
    wd = event->width;
    ht = event->height;

    XCopyArea(XtDisplay(w), w->board.board_pixmap, XtWindow(w),
	w->board.copy_gc, x, y, wd, ht, x, y);
}

static void
DrawGrid(w)
    BoardWidget w;
{
    XSegment segments[2*COLS+ROWS-2+8+4+6*2*8], *sp = segments;
    Position xmin, ymin, xmax, ymax, x, y;
    Dimension spacing, off, len;

    /*
     * Clear Grid Pixmap
     */
    XFillRectangle(XtDisplay(w), w->board.grid_pixmap,
	w->board.bg_gc, 0, 0, w->core.width, w->core.height);
    
    spacing = w->board.spacing;
    xmin = w->board.origin_x;
    ymin = w->board.origin_y;
    xmax = xmin + spacing*(COLS-1);
    ymax = ymin + spacing*(ROWS-1);
    len = spacing/4;
    off = 3;

    /*
     * Draw Frame
     */

    /* top */
    sp->x1 = xmin - 3;
    sp->x2 = xmax + 3;
    sp->y1 = sp->y2 = ymin - 3;
    sp++;
    sp->x1 = xmin - 2;
    sp->x2 = xmax + 2;
    sp->y1 = sp->y2 = ymin - 2;
    sp++;

    /* bottom */
    sp->x1 = xmin - 3;
    sp->x2 = xmax + 3;
    sp->y1 = sp->y2 = ymax + 3;
    sp++;
    sp->x1 = xmin - 2;
    sp->x2 = xmax + 2;
    sp->y1 = sp->y2 = ymax + 2;
    sp++;

    /* left */
    sp->x1 = sp->x2 = xmin - 3;
    sp->y1 = ymin - 3;
    sp->y2 = ymax + 3;
    sp++;
    sp->x1 = sp->x2 = xmin - 2;
    sp->y1 = ymin - 2;
    sp->y2 = ymax + 2;
    sp++;

    /* right */
    sp->x1 = sp->x2 = xmax + 3;
    sp->y1 = ymin - 3;
    sp->y2 = ymax + 3;
    sp++;
    sp->x1 = sp->x2 = xmax + 2;
    sp->y1 = ymin - 2;
    sp->y2 = ymax + 2;
    sp++;

    /*
     * Draw Grid
     */

    /* vertical */
    sp->x1 = sp->x2 = xmin;
    sp->y1 = ymin;
    sp->y2 = ymax;
    sp++;
    sp->x1 = sp->x2 = xmax;
    sp->y1 = ymin;
    sp->y2 = ymax;
    sp++;

    y = ymin + spacing*(ROWS/2-1);
    for (x = xmin + spacing; x < xmax; x += spacing) {
	sp->x1 = sp->x2 = x;
	sp->y1 = ymin;
	sp->y2 = y;
	sp++;
	sp->x1 = sp->x2 = x;
	sp->y1 = y + spacing;
	sp->y2 = ymax;
	sp++;
    }

    /* horizontal */
    for (y = ymax; y >= ymin; y -= spacing) {
	sp->x1 = xmin;
	sp->x2 = xmax;
	sp->y1 = sp->y2 = y;
	sp++;
    }

    /* diagonal */
    sp->x1 = xmin + spacing*3;
    sp->y1 = ymin;
    sp->x2 = xmin + spacing*5;
    sp->y2 = ymin + spacing*2;
    sp++;
    sp->x1 = xmin + spacing*3;
    sp->y1 = ymin + spacing*2;
    sp->x2 = xmin + spacing*5;
    sp->y2 = ymin;
    sp++;
    sp->x1 = xmin + spacing*3;
    sp->y1 = ymax;
    sp->x2 = xmin + spacing*5;
    sp->y2 = ymax - spacing*2;
    sp++;
    sp->x1 = xmin + spacing*3;
    sp->y1 = ymax - spacing*2;
    sp->x2 = xmin + spacing*5;
    sp->y2 = ymax;
    sp++;

    /* Draw Stars */

#define	DRAWANGLE(xx, yy, dx, dy)	\
    sp->x1 = (xx); sp->y1 = (yy);	\
    sp->x2 = (xx)+(dx); sp->y2 = (yy);	\
    sp++;				\
    sp->x1 = (xx); sp->y1 = (yy);	\
    sp->x2 = (xx); sp->y2 = (yy)+(dy);	\
    sp++;

#define	DRAWSTAR(i,j)			\
    x = xmin + spacing*(i);		\
    y = ymin + spacing*(ROWS-1-(j));	\
    DRAWANGLE(x+off, y+off, len, len);	\
    DRAWANGLE(x-off, y+off, -len, len);	\
    DRAWANGLE(x-off, y-off, -len, -len);\
    DRAWANGLE(x+off, y-off, len, -len);

#define	DRAWSTARLEFT(i,j)		\
    x = xmin + spacing*(i);		\
    y = ymin + spacing*(ROWS-1-(j));	\
    DRAWANGLE(x-off, y+off, -len, len);	\
    DRAWANGLE(x-off, y-off, -len, -len);

#define	DRAWSTARRIGHT(i,j)		\
    x = xmin + spacing*(i);		\
    y = ymin + spacing*(ROWS-1-(j));	\
    DRAWANGLE(x+off, y+off, len, len);	\
    DRAWANGLE(x+off, y-off, len, -len);

    /* PAOs */
    DRAWSTAR(1,2);
    DRAWSTAR(7,2);
    DRAWSTAR(1,ROWS-3);
    DRAWSTAR(7,ROWS-3);

    /* ZUs and BINGs */
    DRAWSTARRIGHT(0,3);
    DRAWSTAR(2,3);
    DRAWSTAR(4,3);
    DRAWSTAR(6,3);
    DRAWSTARLEFT(8,3);
    DRAWSTARRIGHT(0,ROWS-4);
    DRAWSTAR(2,ROWS-4);
    DRAWSTAR(4,ROWS-4);
    DRAWSTAR(6,ROWS-4);
    DRAWSTARLEFT(8,ROWS-4);
#undef	DRAWSTAR
#undef	DRAWSTARLEFT
#undef	DRAWSTARRIGHT

    XDrawSegments(XtDisplay(w), w->board.grid_pixmap, w->board.fg_gc,
	segments, sp - segments);
}

static void
DrawPieces(w)
    BoardWidget w;
{
    Position xmin, ymin, xmax, ymax, x, y;
    Dimension spacing, r;
    Piece piece;
    GC gc;
    int i, j;

    /*
     * Copy In Grid Pixmap
     */
    XCopyPlane(XtDisplay(w), w->board.grid_pixmap, w->board.board_pixmap,
	w->board.copy_gc, 0, 0, w->core.width, w->core.height, 0, 0, 1);
    
    spacing = w->board.spacing;
    xmin = w->board.origin_x;
    ymin = w->board.origin_y;
    xmax = xmin + spacing*(COLS-1);
    ymax = ymin + spacing*(ROWS-1);
    r = w->board.radius;

    /*
     * Draw Pieces
     */
    for (x = xmin, i = 0; i < COLS; x += spacing, i++)
    for (y = ymax, j = 0; j < ROWS; y -= spacing, j++) {
	if (IsEmpty(piece = w->board.pieces[i][j]))
	    continue;
	gc = IsBlack(piece) ? w->board.black_gc : w->board.red_gc;
        XSetClipOrigin(XtDisplay(w), gc, x-r, y-r);
	XCopyPlane(XtDisplay(w), w->board.pieces_pixmap[piece-BJU],
	    w->board.board_pixmap, gc, 0, 0, r+r, r+r, x-r, y-r, 1);
    }
}

static void
Resize(w)
    BoardWidget w;
{
    Display *dpy = XtDisplay(w);
    Window win = RootWindowOfScreen(XtScreen(w));
    int dp = DefaultDepthOfScreen(XtScreen(w));
    Dimension wd = w->core.width;
    Dimension ht = w->core.height;

    XFreePixmap(dpy, w->board.board_pixmap);
    XFreePixmap(dpy, w->board.grid_pixmap);
    XFreePixmap(dpy, w->board.temp_pixmap);

    init_size(w);

    w->board.board_pixmap = XCreatePixmap(dpy, win, wd, ht, dp);
    w->board.grid_pixmap = XCreatePixmap(dpy, win, wd, ht, 1);
    w->board.temp_pixmap = XCreatePixmap(dpy, win, wd, ht, dp);

    DrawGrid(w);
    DrawPieces(w);
}

static Boolean
SetValues(w, req, new)
    BoardWidget w, req, new;
{
    Boolean redisplay;

    if (w->board.side != req->board.side) {

	new->board.side = req->board.side;
	init_pieces(new);
	DrawGrid(new);
	DrawPieces(new);

	redisplay = TRUE;
    }

    /*if (w->board.foreground != req->board.foreground ||
	w->board.line_width != req->board.line_width ||
	w->board.font->fid != req->board.font->fid) {*/

    if (w->board.foreground != req->board.foreground ||
	w->board.line_width != req->board.line_width) {

        new->board.foreground = req->board.foreground;
	new->board.line_width = req->board.line_width;
	/*new->board.font->fid = req->board.font->fid;*/

	destroy_gcs(new);
	init_gcs(new);

	redisplay = TRUE;
    }

    return redisplay;
}

static Boolean
LegalMove(w, i, j, ni, nj, real)
    BoardWidget w;
    int i, j, ni, nj;
    Boolean real;
{
    Boolean legal;
    Piece pc, npc;
    int a, b, c;

    if (ni < 0 || ni >= COLS || nj < 0 || nj >= ROWS)
	return FALSE;

    if (i == ni && j == nj)
	return FALSE;

    pc = w->board.pieces[i][j];
    npc = w->board.pieces[ni][nj];

    if (real) {
        if (IsBlack(pc) && (IsBlack(npc) || w->board.turn != BLACK))
	    return FALSE;

        if (IsRed(pc) && (IsRed(npc) || w->board.turn != RED))
	    return FALSE;
    }

    legal = FALSE;

    switch (pc) {

    case BJU:
    case RJU:
	if (i != ni && j != nj)
	    break;
	if (i == ni) {
	    a = min(j,nj); b = max(j,nj); c = 0;
	    for (a++; a < b; a++)
		if (!IsEmpty(w->board.pieces[i][a]))
		    c++;
	} else if (j == nj) {
	    a = min(i,ni); b = max(i,ni); c = 0;
	    for (a++; a < b; a++)
		if (!IsEmpty(w->board.pieces[a][j]))
		    c++;
	}
	legal = (c == 0);
	break;

    case BMA:
    case RMA:
	legal = (abs(i-ni) == 1 && abs(j-nj) == 2 &&
	    IsEmpty(w->board.pieces[i][(j+nj)/2]) ||
	    abs(i-ni) == 2 && abs(j-nj) == 1 &&
	    IsEmpty(w->board.pieces[(i+ni)/2][j]));
	break;

    case BPAO:
    case RPAO:
	if (i != ni && j != nj)
	    break;
	if (i == ni) {
	    a = min(j,nj); b = max(j,nj); c = 0;
	    for (a++; a < b; a++)
		if (!IsEmpty(w->board.pieces[i][a]))
		    c++;
	} else if (j == nj) {
	    a = min(i,ni); b = max(i,ni); c = 0;
	    for (a++; a < b; a++)
		if (!IsEmpty(w->board.pieces[a][j]))
		    c++;
	}
	legal = c == 0 && IsEmpty(npc) || c == 1 && !IsEmpty(npc);
	break;

    case BXIANG:
    case RXIANG:
	if (j < ROWS/2 && nj >= ROWS/2 ||
	    nj < ROWS/2 && j >= ROWS/2)
	    break;
	legal = (abs(i-ni) == 2 && abs(j-nj) == 2 &&
	    IsEmpty(w->board.pieces[(i+ni)/2][(j+nj)/2]));
	break;

    case BSHI:
    case RSHI:
	legal = (ni >= 3 && ni <= 5 && (nj < 3 || nj > ROWS-4) &&
	    abs(i-ni) == 1 && abs(j-nj) == 1);
	break;

    case BJIANG:
    case RSHUAI:
	legal = (ni >= 3 && ni <= 5 && (nj < 3 || nj > ROWS-4) &&
	    (i == ni && abs(j-nj) == 1 || j == nj && abs(i-ni) == 1));
	break;

    case BZU:
    case RBING:
	if (pc == BZU && w->board.side == BLACK ||
	    pc == RBING && w->board.side == RED) {
	    legal = (i == ni && nj == j+1 ||
		(j >= ROWS/2 && j == nj && abs(i-ni) == 1));
	} else {
	    legal = (i == ni && nj == j-1 ||
		(j < ROWS/2 && j == nj && abs(i-ni) == 1));
	}
	break;
    }

    return (legal);
}

static Boolean
Check(w)
    BoardWidget w;
{
    Piece pc;
    Side s;
    int i, j, ki, kj;

    s = w->board.side;
    pc = s == BLACK ? BJIANG : RSHUAI;

    for (kj = 0; kj <= 2; kj++)
    for (ki = 3; ki <= 5; ki++)
	if (w->board.pieces[ki][kj] == pc)
	    goto found;

    return TRUE; /* no KING!? */

found:
    for (i = 0; i < COLS; i++)
    for (j = 0; j < ROWS; j++) {
	pc = w->board.pieces[i][j];
	if (s == BLACK ? IsRed(pc) : IsBlack(pc))
	    if (LegalMove(w, i, j, ki, kj, FALSE))
		return TRUE;
    }
    for (j = ROWS-1; j > ROWS-4; j--) {
	pc = w->board.pieces[ki][j];
	if (s == BLACK ? pc == RSHUAI : pc == BJIANG)
	    break;
    }
    if (j > ROWS-4) {
	for (j--; j > kj; j--)
	    if (!IsEmpty(w->board.pieces[ki][j]))
		return FALSE;
	return TRUE;
    }

    return FALSE;
}

static Position gx, gy;
static int gi, gj;
static int gbtn;

static void
Moving(w, p, event, contp)
    BoardWidget w;
    XtPointer p;
    XMotionEvent *event;
    Boolean *contp;
{
    Display *dpy = XtDisplay(w);
    Dimension r = w->board.radius;
    Position x, y, dx, dy;
    Piece piece;
    Pixmap pixmap;
    GC gc;

    piece = w->board.pieces[gi][gj];
    pixmap = w->board.pieces_pixmap[piece-BJU];

    *contp = FALSE;

    x = event->x; y = event->y;
    dx = x - gx; dy = y - gy;

    if (abs(dx) < 5 && abs(dy) < 5)
	return;

    XCopyArea(dpy, w->board.temp_pixmap, w->board.board_pixmap,
	w->board.copy_gc, gx-r, gy-r, r+r, r+r, gx-r, gy-r);

    gc = IsBlack(piece) ? w->board.black_gc : w->board.red_gc;

    XSetClipOrigin(dpy, gc, x-r, y-r);
    XCopyPlane(dpy, pixmap, w->board.board_pixmap,
	gc, 0, 0, r+r, r+r, x-r, y-r, 1);

    x = min(gx, x);
    y = min(gy, y);

    dx = r+r+abs(dx);
    dy = r+r+abs(dy);

    XCopyArea(dpy, w->board.board_pixmap, XtWindow(w),
	w->board.copy_gc, x-r, y-r, dx, dy, x-r, y-r);

    gx = event->x; gy = event->y;
}

static void
EndMove(w, p, event, contp)
    BoardWidget w;
    XtPointer p;
    XButtonEvent *event;
    Boolean *contp;
{
    Display *dpy = XtDisplay(w);
    Position xmin = w->board.origin_x;
    Position ymin = w->board.origin_y;
    Dimension spacing = w->board.spacing;
    Dimension r = w->board.radius;
    Position x, y, dx, dy;
    Piece piece;
    Pixmap pixmap;
    GC gc;
    Boolean moved;
    int i, j;

    if (event->button != gbtn)
	return;

    *contp = FALSE;

    XtUngrabPointer((Widget)w, CurrentTime);

    piece = w->board.pieces[gi][gj];
    pixmap = w->board.pieces_pixmap[piece-BJU];

    XCopyArea(dpy, w->board.temp_pixmap, w->board.board_pixmap,
	w->board.copy_gc, gx-r, gy-r, r+r, r+r, gx-r, gy-r);

    x = event->x; y = event->y;

    i = (x - xmin + spacing/2) / spacing;
    j = (y - ymin + spacing/2) / spacing;
    j = ROWS-1-j;

    if (LegalMove(w, gi, gj, i, j, TRUE)) {
	BdMovedCallbackStruct cbs;

	cbs.i = gi; cbs.j = gj;
	cbs.ni = i; cbs.nj = j;
	cbs.pc = piece; cbs.npc = w->board.pieces[i][j];

	w->board.pieces[gi][gj] = EMPTY;
	w->board.pieces[i][j] = piece;

	if (w->board.check = Check(w)) {
	    XBell(XtDisplay(w), 10);
	    XBell(XtDisplay(w), 10);
	    XBell(XtDisplay(w), 10);

	    w->board.pieces[gi][gj] = piece;
	    w->board.pieces[i][j] = cbs.npc;
	    i = gi; j = gj;
	} else {
	    w->board.turn = 1 - w->board.turn;
	    XtCallCallbackList((Widget)w, w->board.moved_callbacks,
		(XtPointer)&cbs);
	    w->board.blink_i = i;
	    w->board.blink_j = j;
	    w->board.blink_pc= piece;
	}
    } else {
	i = gi; j = gj;
    }

    x = xmin + spacing*i;
    y = ymin + spacing*(ROWS-1-j);

    gc = IsBlack(piece) ? w->board.black_gc : w->board.red_gc;
    XSetClipOrigin(dpy, gc, x-r, y-r);
    XCopyPlane(dpy, pixmap, w->board.board_pixmap,
	gc, 0, 0, r+r, r+r, x-r, y-r, 1);

    dx = abs(gx-x)+r+r;
    dy = abs(gy-y)+r+r;

    if (x > gx) x = gx;
    if (y > gy) y = gy;

    XCopyArea(dpy, w->board.board_pixmap, XtWindow(w),
	w->board.copy_gc, x-r, y-r, dx, dy, x-r, y-r);

    XtRemoveEventHandler((Widget)w, PointerMotionMask, FALSE, Moving, NULL);
    XtRemoveEventHandler((Widget)w, ButtonReleaseMask, FALSE, EndMove, NULL);

    BlinkOn(w);
}

static void
StartMove(w, event, params, nparams)
    BoardWidget w;
    XButtonEvent *event;
    String *params;
    Cardinal nparams;
{
    Position xmin, ymin;
    Dimension spacing, r;
    Position x, y, dx, dy;
    Piece pc;
    int i, j;

    xmin = w->board.origin_x;
    ymin = w->board.origin_y;
    spacing = w->board.spacing;
    r = w->board.radius;

    i = (event->x - xmin + spacing/2) / spacing;
    j = (event->y - ymin + spacing/2) / spacing;
    j = ROWS-1-j;

    if (i < 0 || j < 0 || i >= COLS || j >= ROWS)
	return;

    if (IsEmpty(pc = w->board.pieces[i][j]))
	return;

    if (IsBlack(pc) && w->board.side != BLACK ||
	IsRed(pc) && w->board.side != RED)
	return;

    x = xmin + spacing*i;
    y = ymin + spacing*(ROWS-1-j);

    dx = event->x - x;
    dy = event->y - y;

    if (dx*dx + dy*dy > r*r)
	return;

    BlinkOff(w);

    XtGrabPointer((Widget)w, FALSE, PointerMotionMask|ButtonReleaseMask,
	GrabModeAsync, GrabModeAsync, XtWindow(w),
	w->board.move_cursor, CurrentTime);

    XCopyArea(XtDisplay(w), w->board.board_pixmap, w->board.temp_pixmap,
	w->board.copy_gc, 0, 0, w->core.width, w->core.height, 0, 0);
    XCopyPlane(XtDisplay(w), w->board.grid_pixmap, w->board.temp_pixmap,
	w->board.copy_gc, x-r, y-r, r+r, r+r, x-r, y-r, 1);

    gi = i; gj = j; gx = x; gy = y; gbtn = event->button;
    XtAddEventHandler((Widget)w, PointerMotionMask, FALSE, Moving, NULL);
    XtAddEventHandler((Widget)w, ButtonReleaseMask, FALSE, EndMove, NULL);
}

Piece
GetPiece(w, i, j)
    BoardWidget w;
    int i, j;
{
    return w->board.pieces[i][j];
}

void
SetPiece(w, i, j, pc)
    BoardWidget w;
    int i, j;
    Piece pc;
{
    Position x, y;
    Dimension r = w->board.radius;
    GC gc;

    w->board.pieces[i][j] = pc;

    BlinkOff(w);

    x = w->board.origin_x + w->board.spacing*i;
    y = w->board.origin_y + w->board.spacing*(ROWS-1-j);

    if (IsEmpty(pc)) {
	XCopyPlane(XtDisplay(w), w->board.grid_pixmap, w->board.board_pixmap,
	    w->board.copy_gc, x-r, y-r, r+r, r+r, x-r, y-r, 1);
    } else {
	gc = IsBlack(pc) ? w->board.black_gc : w->board.red_gc;
	XSetClipOrigin(XtDisplay(w), gc, x-r, y-r);
	XCopyPlane(XtDisplay(w), w->board.pieces_pixmap[pc-BJU],
	    w->board.board_pixmap, gc, 0, 0, r+r, r+r, x-r, y-r, 1);
    }

    if (XtIsRealized((Widget)w))
	XCopyArea(XtDisplay(w), w->board.board_pixmap, XtWindow(w),
	    w->board.copy_gc, x-r, y-r, r+r, r+r, x-r, y-r);

    w->board.blink_i = i;
    w->board.blink_j = j;
    w->board.blink_pc = pc;

    BlinkOn(w);
}

void
MovePiece(w, i, j, ni, nj)
    BoardWidget w;
    int i, j, ni, nj;
{
    Position x, y, nx, ny;
    Dimension r = w->board.radius;
    Piece pc;
    GC gc;

    /*if (!LegalMove(w, i, j, ni, nj, TRUE))
	return;*/

    BlinkOff(w);

    pc = w->board.pieces[i][j];
    w->board.pieces[i][j] = EMPTY;
    w->board.pieces[ni][nj] = pc;
    w->board.turn = 1 - w->board.turn;

    x = w->board.origin_x + w->board.spacing*i;
    y = w->board.origin_y + w->board.spacing*(ROWS-1-j);
    XCopyPlane(XtDisplay(w), w->board.grid_pixmap, w->board.board_pixmap,
	w->board.copy_gc, x-r, y-r, r+r, r+r, x-r, y-r, 1);

    nx = w->board.origin_x + w->board.spacing*ni;
    ny = w->board.origin_y + w->board.spacing*(ROWS-1-nj);

    gc = IsBlack(pc) ? w->board.black_gc : w->board.red_gc;
    XSetClipOrigin(XtDisplay(w), gc, nx-r, ny-r);
    XCopyPlane(XtDisplay(w), w->board.pieces_pixmap[pc-BJU],
	w->board.board_pixmap, gc, 0, 0, r+r, r+r, nx-r, ny-r, 1);

    XCopyArea(XtDisplay(w), w->board.board_pixmap, XtWindow(w),
	w->board.copy_gc, x-r, y-r, r+r, r+r, x-r, y-r);
    XCopyArea(XtDisplay(w), w->board.board_pixmap, XtWindow(w),
	w->board.copy_gc, nx-r, ny-r, r+r, r+r, nx-r, ny-r);

    if (w->board.check = Check(w)) {
	XBell(XtDisplay(w), 10);
	XBell(XtDisplay(w), 10);
	XBell(XtDisplay(w), 10);
    }

    w->board.blink_i = ni;
    w->board.blink_j = nj;
    w->board.blink_pc = pc;

    BlinkOn(w);
}

static void
Blink(w, id)
    BoardWidget w;
    XtIntervalId *id;
{
    GC gc;
    Position x, y;
    Dimension r;
    int interval;

    if (!XtIsRealized((Widget)w) || !w->board.blink_on ||
	IsEmpty(w->board.blink_pc))
    	return;
	
    x = w->board.origin_x + w->board.spacing*w->board.blink_i;
    y = w->board.origin_y + w->board.spacing*(ROWS-1-w->board.blink_j);
    r = w->board.radius;

    if (w->board.blink_show = !w->board.blink_show) {
	gc = IsBlack(w->board.blink_pc) ? w->board.black_gc : w->board.red_gc;
	XSetClipOrigin(XtDisplay(w), gc, x-r, y-r);
	XCopyPlane(XtDisplay(w), w->board.pieces_pixmap[w->board.blink_pc-BJU],
	    w->board.board_pixmap, gc, 0, 0, r+r, r+r, x-r, y-r, 1);
	interval = w->board.blink_interval*3/4;
    } else {
	XCopyPlane(XtDisplay(w), w->board.grid_pixmap, w->board.board_pixmap,
	    w->board.copy_gc, x-r, y-r, r+r, r+r, x-r, y-r, 1);
	interval = w->board.blink_interval/4;
    }
    XCopyArea(XtDisplay(w), w->board.board_pixmap, XtWindow(w),
	w->board.copy_gc, x-r, y-r, r+r, r+r, x-r, y-r);

    w->board.blink_id = XtAppAddTimeOut(
	XtWidgetToApplicationContext((Widget)w),
	interval, Blink, (XtPointer)w);
}

static void
BlinkOn(w)
    BoardWidget w;
{
    if (w->board.blink_on || w->board.blink_interval <= 0 ||
        IsEmpty(w->board.blink_pc))
	return;
    w->board.blink_on = TRUE;
    w->board.blink_show = TRUE;
    Blink(w, NULL);
}

static void
BlinkOff(w)
    BoardWidget w;
{
    if (!w->board.blink_on || w->board.blink_interval <= 0)
	return;

    XtRemoveTimeOut(w->board.blink_id);
    if (!w->board.blink_show) {
	Blink(w, NULL);
	XtRemoveTimeOut(w->board.blink_id);
    }
    w->board.blink_on = FALSE;
}

void
SetBlink(w, i, j, pc)
    BoardWidget w;
    int i, j;
    Piece pc;
{
    BlinkOff(w);

    w->board.blink_i = i;
    w->board.blink_j = j;
    w->board.blink_pc = pc;

    BlinkOn(w);
}
