#ifndef _Board_h
#define _Board_h

/****************************************************************
 *
 * Board widget
 *
 ****************************************************************/

/* Resources:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 background	     Background		Pixel		XtDefaultBackground
 foreground	     Foreground		Pixel		XtDefaultForeground
 blackBackground     Background		Pixel		XtDefaultForeground
 blackForeground     Foreground		Pixel		XtDefaultBackground
 blackBorder	     Foreground		Pixel		XtDefaultForeground
 redBackground	     Background		Pixel		XtDefaultBackground
 redForeground	     Foreground		Pixel		XtDefaultForeground
 redBorder	     Foreground		Pixel		XtDefaultForeground
 borderWidth	     BorderWidth	Dimension	1
 lineWidth	     LineWidth		Dimension	1
 blinkInterval	     BlinkInterval	Integer		1000
 movedCallback	     Callback		Pointer		NULL

*/

/* define any special resource names here that are not in <X11/StringDefs.h> */

#define	XtNlineWidth	"lineWidth"
#define XtNboardFont	"boardFont"
#define	XtNside		"side"
#define	XtNturn		"turn"
#define XtNblackForeground "blackForeground"
#define XtNblackBackground "blackBackground"
#define XtNblackBorder "blackBorder"
#define XtNredForeground "redForeground"
#define XtNredBackground "redBackground"
#define XtNredBorder "redBackorder"
#define XtNblinkInterval "blinkInterval"
#define	XtNmovedCallback "movedCallBack"

#define	XtCLineWidth	"LineWidth"
#define XtCBoardFont	"BoardFont"
#define	XtCSide		"Side"
#define	XtCTurn		"Turn"
#define XtCBlackForeground "BlackForeground"
#define XtCBlackBackground "BlackBackground"
#define XtCBlackBorder "BlackBorder"
#define XtCRedForeground "RedForeground"
#define XtCRedBackground "RedBackground"
#define XtCRedBorder "RedBackorder"
#define XtCBlinkInterval "BlinkInterval"
#define	XtCMovedCallback "MovedCallBack"

/* declare specific BoardWidget class and instance datatypes */

typedef struct _BoardClassRec*	BoardWidgetClass;
typedef struct _BoardRec*	BoardWidget;

/* declare the class constant */

typedef enum {
	EMPTY,
	BJU, BMA, BPAO, BXIANG, BSHI, BJIANG, BZU,
	RJU, RMA, RPAO, RXIANG, RSHI, RSHUAI, RBING,
} Piece;

#define	IsBlack(pc)	((pc) >= BJU && (pc) <= BZU)
#define	IsRed(pc)	((pc) >= RJU && (pc) <= RBING)
#define	IsEmpty(pc)	((pc) == EMPTY)

typedef enum {
	BLACK, RED, OBSERVER
} Side;

#define	COLS	9
#define	ROWS	10

extern WidgetClass boardWidgetClass;

typedef struct
{
    int i, j;
    int ni, nj;
    Piece pc, npc;
} BdMovedCallbackStruct;

extern Piece GetPiece(/*Widget, int, int*/);
extern void SetPiece(/*Widget, int, int, Piece*/);
extern void MovePiece(/*Widget, int, int, int, int*/);
extern void SetBlink(/*Widget, int, int, Piece*/);

#endif /* _Board_h */
