#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "Board.h"

extern int netopen();
extern void netread(), netwrite();

void DataIn(), DataOut();

static void Moved();

struct move_data {
	int i, j, ni, nj;
	Widget w;
	int s;
};

BdMovedCallbackStruct moves[1024];
int nmove = 0;

void quit(), undo();

static char *defaultTranslations =
"Ctrl<Key>U: undo()\n\
 Ctrl<Key>C: quit()\n";

static XtActionsRec actionTable[] = {
        { "undo",       undo },
        { "quit",       quit },
};

main(argc, argv)
	char **argv;
{
	XtAppContext app;
	Widget toplevel, board;
	int s;
	struct move_data rmd, wmd;

	toplevel = XtVaAppInitialize(&app, "CChess", NULL, 0,
		(Cardinal *)&argc, argv, NULL, NULL);

	board = XtVaCreateManagedWidget("board",
		boardWidgetClass, toplevel,
		NULL);

	if (argc > 1) {
		s = netopen(argv[1]);
		XtVaSetValues(board, XtNside, RED, NULL);
	} else {
		s = netopen(NULL);
		XtVaSetValues(board, XtNside, BLACK, NULL);
	}

        XtAppAddActions(app, actionTable, XtNumber(actionTable));
        XtAugmentTranslations(board,
                XtParseTranslationTable(defaultTranslations));

	rmd.w = wmd.w = board;
	rmd.s = wmd.s = s;
	XtAppAddInput(app, s, (XtPointer)XtInputReadMask, netread,
	    (XtPointer)&rmd);
	XtAddCallback(board, XtNmovedCallback, Moved, (XtPointer)&wmd);

	XtRealizeWidget(toplevel);
	XtAppMainLoop(app);
}

void DataIn(md, id)
	struct move_data *md;
	XtInputId id;
{
	BdMovedCallbackStruct *mp;
	Side turn;

	if (md->i < 0) {
		mp = &moves[--nmove];
		SetPiece(md->w, mp->ni, mp->nj, mp->npc);
		SetPiece(md->w, mp->i, mp->j, mp->pc);
		if (nmove <= 0)
		    SetBlink(md->w, 0, 0, EMPTY);
		else {
		    mp = &moves[nmove-1];
		    SetBlink(md->w, mp->ni, mp->nj, mp->pc);
		}
		XtVaGetValues(md->w, XtNturn, &turn, NULL);
		XtVaSetValues(md->w, XtNturn, 1-turn, NULL);
		return;
	}

	mp = &moves[nmove++];
	mp->i = md->i;
	mp->j = md->j;
	mp->ni = md->ni;
	mp->nj = md->nj;
	mp->pc = GetPiece(md->w, md->i, md->j);
	mp->npc = GetPiece(md->w, md->ni, md->nj);

	MovePiece(md->w, md->i, md->j, md->ni, md->nj);
}

void DataOut(md, id)
	struct move_data *md;
	XtInputId id;
{
	XtRemoveInput(id);
}

static void
Moved(w, md, cbs)
	Widget w;
	struct move_data *md;
	BdMovedCallbackStruct *cbs;
{
	XtAppContext app = XtWidgetToApplicationContext(w);
	BdMovedCallbackStruct *mp;
	Side side, turn;
	
	if (cbs == NULL) {	/* a undo move */
		XtVaGetValues(w, XtNside, &side, XtNturn, &turn, NULL);
		if (turn != side || nmove <= 0) {
			XBell(XtDisplay(w), 10);
			return;
		}
		mp = &moves[--nmove];
		SetPiece(w, mp->ni, mp->nj, mp->npc);
		SetPiece(w, mp->i, mp->j, mp->pc);
		if (nmove <= 0)
		    SetBlink(md->w, 0, 0, EMPTY);
		else {
		    mp = &moves[nmove-1];
		    SetBlink(w, mp->ni, mp->nj, mp->pc);
		}
		XtVaSetValues(w, XtNturn, 1-turn, NULL);

		md->i = -1;
		XtAppAddInput(app, md->s, (XtPointer)XtInputWriteMask,
		    netwrite, (XtPointer)md);
		return;
	}

	md->i = COLS-1-cbs->i;
	md->j = ROWS-1-cbs->j;
	md->ni = COLS-1-cbs->ni;
	md->nj = ROWS-1-cbs->nj;

	XtAppAddInput(app, md->s, (XtPointer)XtInputWriteMask, netwrite,
	    (XtPointer)md);

	mp = &moves[nmove++];
	*mp = *cbs;
}

static void
undo(w)
	Widget w;
{
	XtCallCallbacks(w, XtNmovedCallback, NULL);
}

static void
quit()
{
	exit(0);
}
